"use strict";
const GameState = function(pool) {
  this.dynamicObjects = [];
  this.avatars = {};
  this.resizeArrays(pool);
};

GameState.prototype.addPlayer = function(playerId){
	this.avatars[playerId] = {};
};

GameState.prototype.addObject = function(pool, playerId, isAvatar, meshName){
  const i = this.dynamicObjects.length;
  this.dynamicObjects.push({playerId:playerId, isAvatar:isAvatar, meshName:meshName});
  if(isAvatar){
    this.avatars[playerId].objectIndex = i;
  }
  this.resizeArrays(pool);

  this.positions.at(i).setRandom(0, 5);
  this.orientations.at(i).setRandom(-3.14, +3.14);
  this.velocities.at(i).setRandom(-0.1, +0.1);
  this.angVelocities.at(i).setRandom(-0.314, +0.314);
  this.invMasses.at(i).set(1);
  this.invAngMasses.at(i).set(1);
  this.backDrags.at(i).set(0.9);
  this.sideDrags.at(i).set(0.5);
  this.angDrags.at(i).set(0.5);
  this.thrusts.at(i).set(0);
  this.torques.at(i).set(0);
};

GameState.prototype.dropPlayer = function(pool, playerId){
  // remove objects with a certain player id
  this.dynamicObjects = this.dynamicObjects.filter(function(obj){
    return obj.playerId !== playerId;
  });
  // refresh avatar indices
  delete this.avatars[playerId];
  for(let i=0; i<this.dynamicObjects.length; i++){
    const obj = this.dynamicObjects[i];
    if(obj.isAvatar){
      this.avatars[obj.playerId].objectIndex = i;
    }
  }
  this.resizeArrays(pool);
};

GameState.prototype.update = function(dt) {
  this.aheads.cossin(this.orientations);
  this.forces.mulWithVec1s(this.aheads, this.thrusts);
  this.velocities.addScaled(this.velocities, this.forces, dt);
  this.angVelocities.addScaled(this.angVelocities, this.torques, dt);
  this.positions.addScaled(this.positions, this.velocities, dt);
  this.orientations.addScaled(this.orientations, this.angVelocities, dt);
  this.aheadSpeeds.dotVec2s(this.velocities, this.aheads);
  this.aheadVelocities.mulWithVec1s(this.aheads, this.aheadSpeeds);
  this.sideVelocities.sub(this.velocities, this.aheadVelocities);
  this.backDragFactors.exp(this.backDrags, dt);
  this.aheadVelocities.mulWithVec1s(this.aheadVelocities, this.backDragFactors);
  this.sideDragFactors.exp(this.sideDrags, dt);
  this.sideVelocities.mulWithVec1s(this.sideVelocities, this.sideDragFactors);  
  this.velocities.add(this.aheadVelocities, this.sideVelocities);
  this.angDragFactors.exp(this.angDrags, dt);
  this.angVelocities.mul(this.angVelocities, this.angDragFactors);  

  const relativeVelocity = new Vec2();
  let diff = new Vec2();
  for(let i=0; i<this.dynamicObjects.length; i++) {
    for(let j=i+1; j<this.dynamicObjects.length; j++) {
      diff.set(this.positions.at(i)).sub(this.positions.at(j));
      const dist2 = diff.dot(diff);
      if(dist2 < 4) {
        diff.mul( 1.0 / Math.sqrt(dist2) );
        this.positions.at(i).addScaled(0.05, diff);
        this.positions.at(j).addScaled(-0.05, diff);
        let tangent = new Vec2(-diff.y, diff.x);
        let vi = this.velocities.at(i);
        let bi = this.angVelocities.at(i);
        let vj = this.velocities.at(j);
        let bj = this.angVelocities.at(j);
        relativeVelocity.set(vi).sub(vj).addScaled(bi.x - bj.x, tangent);
        const impulseLength = diff.dot(relativeVelocity);
        diff.mul( impulseLength * 1.5 /*restitution*/ );
        const frictionLength = tangent.dot(relativeVelocity);        
        tangent.mul(frictionLength * 0.5 /*friction*/);
        vi.sub(diff).sub(tangent);
        vj.add(diff).add(tangent);
        bi.add(frictionLength /* *radius*/ );
        bj.add(frictionLength /* *radius*/ );        
      }
    }
    const pi = this.positions.at(i);
    diff.set((pi.x + 8000) % 8, (pi.y + 8000) % 8);
    diff.sub(4, 4);
    const dist2 = diff.dot(diff);
    if(dist2 < 4) {
      // LABTODO: ütközésválasz statikusokra
    }
  }
};

GameState.prototype.resizeArrays = function(pool) {
  this.positions     = pool.positions    .subarray(0, this.dynamicObjects.length);
  this.orientations  = pool.orientations .subarray(0, this.dynamicObjects.length);
  this.velocities    = pool.velocitys    .subarray(0, this.dynamicObjects.length);
  this.angVelocities = pool.angVelocitys .subarray(0, this.dynamicObjects.length);
  this.invMasses     = pool.invMasss     .subarray(0, this.dynamicObjects.length);
  this.invAngMasses  = pool.invAngMasss  .subarray(0, this.dynamicObjects.length);
  this.backDrags     = pool.backDrags    .subarray(0, this.dynamicObjects.length);
  this.sideDrags     = pool.sideDrags    .subarray(0, this.dynamicObjects.length);
  this.angDrags      = pool.angDrags     .subarray(0, this.dynamicObjects.length);
  this.thrusts       = pool.thrusts      .subarray(0, this.dynamicObjects.length);
  this.torques       = pool.torques      .subarray(0, this.dynamicObjects.length);

  this.forces          = pool.forces         .subarray(0, this.dynamicObjects.length);
  this.aheads          = pool.aheads         .subarray(0, this.dynamicObjects.length);
  this.aheadSpeeds     = pool.aheadSpeeds    .subarray(0, this.dynamicObjects.length);
  this.aheadVelocities = pool.aheadVelocitys .subarray(0, this.dynamicObjects.length);
  this.sideVelocities  = pool.sideVelocitys  .subarray(0, this.dynamicObjects.length);  
  this.backDragFactors = pool.backDragFactors.subarray(0, this.dynamicObjects.length);
  this.sideDragFactors = pool.sideDragFactors.subarray(0, this.dynamicObjects.length);
  this.angDragFactors  = pool.angDragFactors .subarray(0, this.dynamicObjects.length);  
};

GameState.prototype.set = function(serverState) {
  if(this.dynamicObjects.length !== serverState.dynamicObjects.length) {
    return;
  }
  for(let i=0; i<this.positions.storage.length; i++) {
    this.positions    .storage[i] = serverState.positions    .storage[i];
    this.velocities   .storage[i] = serverState.velocities   .storage[i];
  }
  for(let i=0; i<this.orientations.storage.length; i++) {
    this.orientations .storage[i] = serverState.orientations .storage[i];
    this.angVelocities.storage[i] = serverState.angVelocities.storage[i];
    this.invMasses    .storage[i] = serverState.invMasses    .storage[i];
    this.invAngMasses .storage[i] = serverState.invAngMasses .storage[i];
    this.backDrags    .storage[i] = serverState.backDrags    .storage[i];
    this.sideDrags    .storage[i] = serverState.sideDrags    .storage[i];
    this.angDrags     .storage[i] = serverState.angDrags     .storage[i];
    this.thrusts      .storage[i] = serverState.thrusts      .storage[i];
    this.torques      .storage[i] = serverState.torques      .storage[i];
  }
  this.avatars = serverState.avatars;
};

if (typeof module !== 'undefined' && typeof module.exports !== 'undefined'){
  module.exports = GameState;
}

